"""
@Project  : dichotomous-score
@File     : setup.py
@Author   : Shaobo Cui
@Date     : 17.10.2024 14:16
"""

# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os

def get_version():
    with open("oppositescore/__init__.py", encoding="utf-8") as f:
        for line in f:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Version not found")

# Read the README file for the long description
with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read the requirements file
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), encoding='utf-8') as f:
    requirements = [line.strip() for line in f if line.strip()]

setup(
    include_package_data=True,
    name='opposite-score',
    version=get_version(),
    description='Optimized Text Embeddings Designed for Measuring Opposite/Contrasting Relationships',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Shaobo Cui',
    author_email='shaobo.cui@epfl.ch',
    url='https://github.com/cui-shaobo/conditional-dichotomy-quantification',
    packages=find_packages(include=['oppositescore', 'oppositescore.*']),
    install_requires=requirements,
    zip_safe=False,
    keywords='text embedding, NLP, opposite relationships, dichotomy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.7, <3.12',
)


"""
@Project  : dichotomous-score
@File     : dichotomye_trainer.py
@Author   : Shaobo Cui
@Date     : 08.09.2024 16:11
"""
from typing import Optional, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers import Trainer, AutoModel


from oppositescore.model.angle import Pooler, DatasetFormats, check_llm, get_pooling
from oppositescore.trainer.loss import DichotomyLoss
from oppositescore.utils.utils import logger


class DichotomyTrainer(Trainer):
    """
    Custom Huggingface Trainer for DichotomyE.

    :param pooler: Pooler. Required
    :param loss_kwargs: Optional[Dict]. Default None.
    :param dataset_format: str. Default DatasetFormats.A
    :param teacher_name_or_path: Optional[str]. For distribution alignment.
    :param **kwargs: other parameters of Trainer.
    """

    def __init__(self,
                 pooler: Pooler,
                 loss_kwargs: Optional[Dict] = None,
                 dataset_format: str = DatasetFormats.A,
                 teacher_name_or_path: Optional[str] = None,
                 teacher_pooling_strategy: str = 'cls',
                 **kwargs):
        super().__init__(**kwargs)
        self.pooler = pooler
        if loss_kwargs is None:
            loss_kwargs = {}
        self.loss_fct = DichotomyLoss(dataset_format=dataset_format, **loss_kwargs)
        self.teacher_name_or_path = teacher_name_or_path
        self.teacher_pooling_strategy = teacher_pooling_strategy
        if teacher_name_or_path is not None:
            logger.info('Teacher detected! '
                        'please ensure the teacher has the same tokenizer as the backbone model!')
            assert not check_llm(teacher_name_or_path), ('Currently not support LLMs alignment,'
                                                         f' teacher={teacher_name_or_path}')
            teacher_backbone = AutoModel.from_pretrained(
                teacher_name_or_path,
                trust_remote_code=True,
                torch_dtype=self.pooler.model.dtype).to(self.pooler.model.device)

            self.teacher_pooler = Pooler(
                teacher_backbone,
                pooling_strategy=self.teacher_pooling_strategy,
                padding_strategy=self.pooler.padding_strategy)
            logger.info(f'Train with teacher={teacher_name_or_path}')

    def distillation_loss(self,
                          inputs: torch.Tensor,
                          targets: torch.Tensor,
                          mse_weight: float = 1.0,
                          kl_temperature: float = 1.0) -> torch.Tensor:
        """ Compute distillation loss.

        :param inputs: torch.Tensor. Input tensor.
        :param targets: torch.Tensor. Target tensor.
        :param mse_weight: float. MSE weight. Default 1.0.
        :param kl_temperature: float. KL temperature. Default 1.0.

        :return: torch.Tensor. Distillation loss.
        """
        loss = 0.
        if mse_weight > 0:
            loss += mse_weight * nn.MSELoss()(inputs, targets)
        if kl_temperature > 0:
            loss += nn.KLDivLoss(reduction='batchmean')(
                F.log_softmax(inputs / kl_temperature, dim=-1),
                F.softmax(targets / kl_temperature, dim=-1)
            ) * kl_temperature
        return loss

    # Rewrite the compute_loss for training dichotomy embedding.
    def compute_loss(self, model, inputs, return_outputs=False):
        """ Compute loss for AnglE.

        :param model: Huggingface model.
        :param inputs: Dict. Model inputs.
        :param return_outputs: bool. Return outputs or not. Default False.

        :return: torch.Tensor. Loss.
        """
        labels = inputs.pop("labels", None)
        if self.teacher_name_or_path is not None:
            all_outputs = self.pooler(inputs, layer_index=-1, return_all_layer_outputs=True)[-1]
            outputs = get_pooling(all_outputs, inputs,
                                  self.pooler.pooling_strategy,
                                  self.pooler.padding_strategy)
            loss = self.loss_fct(labels, outputs)
            with torch.no_grad():
                self.teacher_pooler.model = self.teacher_pooler.model.to(self.pooler.model.device)
                align_outputs = self.teacher_pooler(inputs)

            alignment_loss = self.distillation_loss(
                all_outputs if self.teacher_pooling_strategy == 'all' else outputs,
                align_outputs,
                mse_weight=0.0,
                kl_temperature=1.0)
            loss += alignment_loss
        else:
            outputs = self.pooler(inputs)
            loss = self.loss_fct(labels, outputs)
        # print("#" * 20, labels)
        loss.requires_grad = True
        return (loss, outputs) if return_outputs else loss




"""
@Project  : dichotomous-score
@File     : dichotomyInferencer.py
@Author   : Shaobo Cui
@Date     : 17.10.2024 20:42
"""


from typing import List
import numpy as np
from boltons.iterutils import chunked_iter
from tqdm import tqdm
from sklearn.metrics.pairwise import paired_cosine_distances
from oppositescore.model.base import AngleBase


class DichotomyInferencer(object):
    def __init__(
        self,
        context: List[str],
        sent1: List[str],
        sent2: List[str],
        batch_size: int = 1
    ):
        assert len(context) == len(sent1) == len(sent2), \
            "context, positive, negative, and neutral must have the same length"
        self.context = context
        self.sent1 = sent1
        self.sent2 = sent2
        self.batch_size = batch_size

    def __call__(self, model: AngleBase, show_progress: bool = True, **kwargs) -> dict:
        """ Evaluate the model on the given dataset using DCF.
        :param model: DichotomyE, the model to evaluate.
        :param show_progress: bool, whether to show a progress bar during evaluation.
        :param kwargs: Additional keyword arguments to pass to the `encode` method of the model.
        :return: dict, The evaluation results including DCF.
        """
        # To store embeddings for context-aware combinations
        context_sent1_embeddings = []
        context_sent2_embeddings = []

        # Process data in batches
        for chunk in tqdm(chunked_iter(range(len(self.context)), self.batch_size),
                          total=len(self.context) // self.batch_size,
                          disable=not show_progress):
            # Concatenate context with positive, negative, and neutral respectively
            batch_context_sent1 = [self.context[i] + " " + self.sent1[i] for i in chunk]
            batch_context_sent2 = [self.context[i] + " " + self.sent2[i] for i in chunk]

            # Encode the concatenated texts
            context_sent1_embeddings.append(model.encode(batch_context_sent1, **kwargs))
            context_sent2_embeddings.append(model.encode(batch_context_sent2, **kwargs))

        # Concatenate the embeddings along the batch dimension
        context_sent1_embeddings = np.concatenate(context_sent1_embeddings, axis=0)
        context_sent2_embeddings = np.concatenate(context_sent2_embeddings, axis=0)

        # Calculate angles for DCF
        angles_pos_neg = self._compute_angles(context_sent1_embeddings, context_sent2_embeddings)

        return angles_pos_neg

    @staticmethod
    def _compute_angles(embeddings1, embeddings2):
        """ Compute angles between pairs of embeddings.
        :param embeddings1: np.array, embeddings of the first set.
        :param embeddings2: np.array, embeddings of the second set.
        :return: np.array, angles between the embeddings.
        """
        angles = paired_cosine_distances(embeddings1, embeddings2)
        # print(cos_sim)
        return angles